﻿using MapTo.Sources;
using static MapTo.Sources.Constants;
using System;
using System.Collections.Generic;
using System.Text;
using System.Collections.Immutable;
using System.Diagnostics;
using Microsoft.CodeAnalysis;
#pragma warning disable CS8602

namespace MapTo.Extensions
{
    internal static class CommonSource
    {
        internal static SourceCode GenerateStructOrClass(this MappingModel model, string structOrClass)
        {
            const bool writeDebugInfo = false;
            
            List<string> constructorHeaders = new List<string>();

            using var builder = new SourceBuilder()
                .WriteLine(GeneratedFilesHeader)
                .WriteNullableContextOptionIf(model.Options.SupportNullableReferenceTypes)
                .WriteUsings(model.Usings)
                .WriteLine()
                // Namespace declaration
                .WriteLine($"namespace {model.Namespace}")
                .WriteOpeningBracket();

            builder
                // Class declaration
                .WriteLine($"partial {structOrClass} {model.TypeIdentifierName}")
                .WriteOpeningBracket()
                .WriteLine()
                // Class body
                .GeneratePublicEmptyConstructor(model, ref constructorHeaders, true)
                .GeneratePublicConstructor(model, ref constructorHeaders)
                .GeneratePublicConstructor(model, ref constructorHeaders, true);
            
            
            if (model.IsTypeUpdatable) builder.GenerateUpdateMethod(model);
            if (model.IsJsonExtension) builder.WriteToJsonMethod(model);


            builder
                .WriteLine()
                // End class declaration
                .WriteClosingBracket()
                .WriteLine()
                // End namespace declaration
                .WriteClosingBracket();

            return new(builder.ToString(), $"{model.Namespace}.{model.TypeIdentifierName}.g.cs");
        }

        private static SourceBuilder GeneratePublicConstructor(this SourceBuilder builder, MappingModel model, ref List<string> constructorHeaders, bool filterNonMapped = false)
        {
            const string mappingContextParameterName = "context";

            foreach (var targetSourceType in model.MappedSourceTypes)
            {
                var sourceClassParameterName = targetSourceType.SourceTypeIdentifierName.ToCamelCase();
                var baseConstructor = /*model.HasMappedBaseClass ? $" : base({mappingContextParameterName}, {sourceClassParameterName})" :*/ string.Empty;
                var stringBuilder = new StringBuilder();
                var otherProperties = new List<MappedMember>();

                foreach (var property in targetSourceType.TypeProperties)
                {
                    if (!targetSourceType.SourceProperties.IsMappedProperty(property) && !filterNonMapped)
                    {
                        stringBuilder.Append(", ");
                        stringBuilder.Append($"{property.FullyQualifiedType} {property.SourcePropertyName.ToCamelCase()}");
                        otherProperties.Add(property);
                    }
                    
                }
                foreach (var property in targetSourceType.TypeFields)
                {
                    if (!targetSourceType.SourceFields.IsMappedProperty(property) && !filterNonMapped)
                    {
                        stringBuilder.Append(", ");
                        stringBuilder.Append($"{property.FullyQualifiedType} {property.SourcePropertyName.ToCamelCase()}");
                        otherProperties.Add(property);
                    }
                }
                
                var readOnlyPropertiesArguments = stringBuilder.ToString();

                var constructorHeader =
                    $"public {model.TypeIdentifierName}({targetSourceType.SourceType} {sourceClassParameterName}{readOnlyPropertiesArguments}){baseConstructor}";


                bool hasAlreadyConstructor = false;
                
                foreach (var header in constructorHeaders)
                {
                    if(constructorHeader.Contains(header)) hasAlreadyConstructor = true;
                }

                if (hasAlreadyConstructor) continue;

                constructorHeaders.Add(constructorHeader);
                
                builder
                    .WriteLine(constructorHeader)
                    .WriteOpeningBracket()
                    .WriteAssignmentMethod(model, filterNonMapped ? null : otherProperties.ToArray().ToImmutableArray(), sourceClassParameterName, mappingContextParameterName, filterNonMapped);

                builder.WriteClosingBracket()
                    .WriteLine();
            }

            // End constructor declaration
            return builder;
        }

        private static SourceBuilder GeneratePublicEmptyConstructor(this SourceBuilder builder, MappingModel model, ref List<string> constructorHeaders, bool filterNonMapped = false)
        {
            const string mappingContextParameterName = "context";

            foreach (var targetSourceType in model.MappedSourceTypes)
            {
                var baseConstructor = /*model.HasMappedBaseClass ? $" : base({mappingContextParameterName}, {sourceClassParameterName})" :*/ string.Empty;
                
                var constructorHeader =
                    $"public {model.TypeIdentifierName}(){baseConstructor}";


                bool hasAlreadyConstructor = false;
                
                foreach (var header in constructorHeaders)
                {
                    if(constructorHeader.Contains(header)) hasAlreadyConstructor = true;
                }

                if (hasAlreadyConstructor) continue;

                constructorHeaders.Add(constructorHeader);
                
                builder
                    .WriteLine(constructorHeader)
                    .WriteOpeningBracket()
                    .WriteClosingBracket()
                    .WriteLine();
            }

            // End constructor declaration
            return builder;
        }

        private static bool IsMappedProperty(this System.Collections.Immutable.ImmutableArray<MappedMember> properties, MappedMember property)
        {

            foreach (var prop in properties)
            {
                if (prop.Name == property.Name) return true;
            }

            return false;
        }

        private static SourceBuilder WriteToJsonMethod(this SourceBuilder builder, MappingModel model)
        {
            builder
                .WriteLine($"public string ToJson()")
                .WriteOpeningBracket()
                .WriteLine("var stringBuilder = new System.Text.StringBuilder();")
                .WriteLine(GetStringBuilderAppendNoInterpolation("{"));

            foreach (var targetSourceType in model.MappedSourceTypes)
            {
                foreach (var property in targetSourceType.TypeProperties)
                {
                    if (!property.isEnumerable)
                        HandlePropertyEnumerable(builder, property);
                    else
                    {
                        builder = WriteJsonField(builder, property);
                    }
                }
                foreach (var property in targetSourceType.TypeFields)
                {
                    if (!property.isEnumerable)
                        HandleFieldEnumerable(builder, property);
                    else
                    {
                        builder.WriteLine(GetStringBuilderAppend($"\\\"{property.Name.ToCamelCase()}\\\" : [{GetJsonArrayValue(property, ref builder)}],"));
                    }
                }
                
                builder.WriteLine(GetStringBuilderAppendNoInterpolation("}"));
                builder.WriteLine("return stringBuilder.ToString();");
                builder.WriteClosingBracket();
            }
            

            return builder;
        }

        private static SourceBuilder WriteJsonField(SourceBuilder builder, MappedMember property)
        {
            builder.WriteLine(
                GetStringBuilderAppend(
                    $"\\\"{property.Name.ToCamelCase()}\\\" : [{GetJsonArrayValue(property, ref builder)}],"));
            return builder;
        }

        private static void HandleEnumerable(SourceBuilder builder, MappedMember property)
        {
            var symbol = property.ActualSymbol as IPropertySymbol;
#pragma warning disable CS8602
            builder.WriteCommentArray(symbol.Parameters, nameof(symbol.Parameters));
#pragma warning restore CS8602
            builder.WriteCommentArray(symbol.TypeCustomModifiers, nameof(symbol.TypeCustomModifiers));

            builder.WriteComment($"Is enumerable {(property.ActualSymbol as IPropertySymbol).Parameters}");
            builder.WriteLine(
                GetStringBuilderAppend($"\\\"{property.Name.ToCamelCase()}\\\" : {GetJsonValue(property, builder)},"));
        }

        
        private static void HandleFieldEnumerable(SourceBuilder builder, MappedMember property)
        {
            HandleEnumerable(builder, property);
        }

        private static void HandlePropertyEnumerable(SourceBuilder builder, MappedMember property)
        {
            HandleEnumerable(builder, property);
        }

        private static string GetJsonArrayValue(MappedMember member, ref SourceBuilder builder)
        {
            if (member.isEnumerable)
            {
                // get underlying type (check if is a json extension)

                builder.WriteLine("var arrStrBuilder = new StringBuilder();");
                
                foreach (var named in member.NamedTypeSymbol?.TypeArguments!)
                {
                    bool? containedTypeIsJsonEXtension = named?.HasAttribute(MappingContext.JsonExtensionAttributeSymbol);
                    if (!containedTypeIsJsonEXtension.HasValue) continue;
                    builder.WriteLine($"foreach (var v in {member.SourcePropertyName.ToString()})");
                    builder.WriteOpeningBracket();
                    builder.WriteLine("arrStrBuilder.Append(v.ToJson());");
                    builder.WriteLine("arrStrBuilder.Append(\", \");");
                    builder.WriteClosingBracket();
                }
                builder.WriteLine("arrStrBuilder.Remove(arrStrBuilder.Length -1, 1);"); 
            }

            return "{arrStrBuilder.ToString()}";
        }
        private static string GetJsonValue(MappedMember member, SourceBuilder builder)
        {

            if (member.FullyQualifiedType == "string") return $"\\\"{{{member.SourcePropertyName}}}\\\"";
            if (member.FullyQualifiedType is "int" or "double" or "float" or "long") return $"{{{member.SourcePropertyName}}}";

            return "";
        }

        private static string GetStringBuilderAppend(string stringToAppend)
        {
            return $"stringBuilder.Append($\"{stringToAppend}\");";
        }
        private static string GetStringBuilderAppendNoInterpolation(string stringToAppend)
        {
            return $"stringBuilder.Append(\"{stringToAppend}\");";
        }

        private static SourceBuilder WriteAssignmentMethod(this SourceBuilder builder, MappingModel model, System.Collections.Immutable.ImmutableArray<MappedMember>? otherProperties,
            string? sourceClassParameterName, string mappingContextParameterName, bool fromUpdate)
        {

            List<MappedMember> _addedMembers = new List<MappedMember>();

            foreach (var targetSourceType in model.MappedSourceTypes)
            {
                foreach (var property in targetSourceType.SourceProperties)
                {
                    if (property.isReadOnly && fromUpdate) continue;
                    if(_addedMembers.Contains(property)) continue;
                    

                    builder.WriteLine($"{property.Name} = {sourceClassParameterName}.{property.SourcePropertyName};");
                    _addedMembers.Add(property);

                }

                foreach (var property in targetSourceType.SourceFields)
                {
                    if (property.isReadOnly && fromUpdate) continue;
                    if(_addedMembers.Contains(property)) continue;

                    builder.WriteLine($"{property.Name} = {sourceClassParameterName}.{property.SourcePropertyName};");
                    _addedMembers.Add(property);

                }

                if (otherProperties == null) return builder;

                foreach (var property in otherProperties)
                {
                    if(_addedMembers.Contains(property)) continue;

                    builder.WriteLine(property.MappedSourcePropertyTypeName is null
                        ? $"{property.Name} = {property.SourcePropertyName.ToCamelCase()};"
                        : "");
                    _addedMembers.Add(property);


                }
            }

            

            return builder;

        }


        private static SourceBuilder GenerateUpdateMethod(this SourceBuilder builder, MappingModel model)
        {

            foreach (var targetSourceType in model.MappedSourceTypes)
            {
                var sourceClassParameterName = targetSourceType.SourceTypeIdentifierName.ToCamelCase();

                builder
                    .GenerateUpdaterMethodsXmlDocs(model, sourceClassParameterName, targetSourceType)
                    .WriteLine($"public void Update({targetSourceType.SourceType} {sourceClassParameterName})")
                    .WriteOpeningBracket()
                    .WriteAssignmentMethod(model, null, sourceClassParameterName, "context", true)
                    .WriteClosingBracket()
                    .WriteLine();
            }
  

            return builder;
        }

        private static SourceBuilder GenerateUpdaterMethodsXmlDocs(this SourceBuilder builder, MappingModel model, string sourceClassParameterName,
            MappedSourceType mappedSourceType)
        {
            if (!model.Options.GenerateXmlDocument)
            {
                return builder;
            }
            builder
                .WriteLine("/// <summary>")
                .WriteLine($"/// Updates <see cref=\"{model.TypeIdentifierName}\"/> and sets its participating properties")
                .WriteLine($"/// using the property values from <paramref name=\"{sourceClassParameterName}\"/>.")
                .WriteLine("/// </summary>")
                .WriteLine($"/// <param name=\"{sourceClassParameterName}\">The instance of <see cref=\"{mappedSourceType.SourceType}\"/> to use as source.</param>");
            return builder;
        }
        
        private static SourceBuilder GenerateEnumerableJsonSourceTypeExtensionMethod(this SourceBuilder builder, MappingModel model)
        {

            foreach (var targetSourceType in model.MappedSourceTypes)
            {
                var sourceClassParameterName = targetSourceType.SourceTypeIdentifierName.ToCamelCase();

                builder
                    .WriteLineIf(model.Options.SupportNullableStaticAnalysis, $"[return: NotNullIfNotNull(\"{sourceClassParameterName}\")]")
                    .WriteLine($"{model.Options.GeneratedMethodsAccessModifier.ToLowercaseString()} static string ToJson(this IEnumerable<{targetSourceType.SourceType}{model.Options.NullableReferenceSyntax}> {sourceClassParameterName}List)")
                    .WriteOpeningBracket()
                    .WriteLine($"return {sourceClassParameterName} == null ? null : new {model.TypeIdentifierName}({sourceClassParameterName});")
                    .WriteClosingBracket();
            }


            return builder;
        }
    }
}
