﻿using System;
using System.Collections.Immutable;
using MapTo.Extensions;
using Microsoft.CodeAnalysis;
using Microsoft.CodeAnalysis.CSharp;

namespace MapTo
{
    internal enum AccessModifier
    {
        Public,
        Internal,
        Private
    }

    internal enum NullStaticAnalysisState
    {
        Default,
        Enabled,
        Disabled
    }

    internal record SourceCode(string Text, string HintName);

    internal record MappedMember(
        string Name,
        string FullyQualifiedType,
        string Type,
        string? TypeConverter,
        ImmutableArray<string> TypeConverterParameters,
        string SourcePropertyName,
        string? MappedSourcePropertyTypeName,
        string? EnumerableTypeArgument,
        ISymbol ActualSymbol,
        INamedTypeSymbol? NamedTypeSymbol,
        bool isEnumerable,
        bool isReadOnly)
    {
        public bool IsEnumerable => EnumerableTypeArgument is not null;
    }

    internal record MappedSourceType
    (
        string SourceNamespace,
        string SourceTypeIdentifierName,
        string SourceTypeFullName,
        ImmutableArray<MappedMember> SourceProperties,
        ImmutableArray<MappedMember> SourceFields,
        ImmutableArray<MappedMember> TypeProperties,
        ImmutableArray<MappedMember> TypeFields,
        bool GenerateSecondaryConstructor
    )
    {
        public string SourceType => SourceTypeFullName;
    }


    internal record MappingModel(
        SourceGenerationOptions Options,
        string? Namespace,
        SyntaxTokenList Modifiers,
        string Type,
        string TypeIdentifierName,
        bool IsTypeUpdatable,
        bool IsJsonExtension,
        ImmutableArray<MappedSourceType> MappedSourceTypes,
        bool HasMappedBaseClass,
        ImmutableArray<string> Usings
    );

    internal interface IEfMethodsModel { }




    internal record EfMethodsModel(
        SourceGenerationOptions Options,
        string Namespace,
        string ContextTypeName,
        string ContextFullType,
        ImmutableArray<EfEntityDataModel> MethodsModels,
        ImmutableArray<string> Usings
    );

    internal class EfEntityDataModel
    {
        public string PropertyName { get; set; }
        public string EntityTypeFullName { get; set; }
        public string EntityTypeIdentifierName { get; set; }

        public EfEntityDataModel(string propertyName, string entityTypeFullName, string entityTypeIdentifierName)
        {
            PropertyName = propertyName;
            EntityTypeFullName = entityTypeFullName;
            EntityTypeIdentifierName = entityTypeIdentifierName;
        }
    }


    internal class EfAddMethodsModel : EfEntityDataModel
    {
        public string CreateTypeFullName { get; set; }
        public string CreateTypeIdentifierName { get; set; }
        public string ReturnTypeFullName { get; set; }
        public string ReturnTypeIdentifierName { get; set; }

        public EfAddMethodsModel(EfEntityDataModel entity, string createTypeFullName,
        string createTypeIdentifierName,
        string returnTypeFullName,
        string returnTypeIdentifierName) : base(entity.PropertyName, entity.EntityTypeFullName, entity.EntityTypeIdentifierName)
        {
            CreateTypeFullName = createTypeFullName;
            CreateTypeIdentifierName = createTypeIdentifierName;
            ReturnTypeIdentifierName = returnTypeIdentifierName;
            ReturnTypeFullName = returnTypeFullName;
        }
    }

    internal class EfGetOneByModel : EfEntityDataModel
    {
        public string ByParamPropertyName { get; set; }
        public string ByParamFullTypeName { get; set; }
        public string ByParamTypeName { get; set; }
        public string ReturnTypeIdentifierName { get; set; }
        public string ReturnTypeFullName { get; set; }


        public EfGetOneByModel(EfEntityDataModel entity, string byParamPropertyName,
        string byParamFullTypeName,
        string returnTypeFullName,
        string returnTypeIdentifierName) : base(entity.PropertyName, entity.EntityTypeFullName, entity.EntityTypeIdentifierName)
        {
            ByParamPropertyName = byParamPropertyName;
            ByParamFullTypeName = byParamFullTypeName;
            ReturnTypeIdentifierName = returnTypeIdentifierName;
            ReturnTypeFullName = returnTypeFullName;
            ByParamTypeName = returnTypeFullName;
        }
    }
    
    internal class EfGetOneWithModel : EfEntityDataModel
    {
        public string ReturnTypeIdentifierName { get; set; }
        public string ReturnTypeFullName { get; set; }

        public EfGetOneWithModel(EfEntityDataModel entity, string returnTypeFullName,
        string returnTypeIdentifierName) : base(entity.PropertyName, entity.EntityTypeFullName, entity.EntityTypeIdentifierName)
        {
            ReturnTypeIdentifierName = returnTypeIdentifierName;
            ReturnTypeFullName = returnTypeFullName;
        }
    }
        
    internal class EfGetManyModel : EfEntityDataModel
    {
        public string ReturnTypeIdentifierName { get; set; }
        public string ReturnTypeFullName { get; set; }

        public EfGetManyModel(EfEntityDataModel entity, string returnTypeFullName,
        string returnTypeIdentifierName) : base(entity.PropertyName, entity.EntityTypeFullName, entity.EntityTypeIdentifierName)
        {
            ReturnTypeIdentifierName = returnTypeIdentifierName;
            ReturnTypeFullName = returnTypeFullName;
        }
    }


    internal class EfUpdateMethodsModel : EfEntityDataModel
    {
        public string UpdateTypeFullName;
        public string UpdateTypeIdentifierName;
        public  string ReturnTypeFullName;
        public string ReturnTypeIdentifierName;
        public string KeyPropertyName;
        public string KeyFullTypeName;
        public EfUpdateMethodsModel(EfEntityDataModel entity,
        string updateTypeFullName,
        string updateTypeIdentifierName,
        string returnTypeFullName,
        string returnTypeIdentifierName,
        string keyPropertyName,
        string keyFullTypeName) : base(entity.PropertyName, entity.EntityTypeFullName, entity.EntityTypeIdentifierName)
        {
            UpdateTypeFullName = updateTypeFullName;
            UpdateTypeIdentifierName = updateTypeIdentifierName;
            ReturnTypeFullName= returnTypeFullName;
            ReturnTypeIdentifierName = returnTypeIdentifierName;
            KeyPropertyName = keyPropertyName;  
            KeyFullTypeName = keyFullTypeName;
        }
    }

    internal record SourceGenerationOptions(
        AccessModifier ConstructorAccessModifier,
        AccessModifier GeneratedMethodsAccessModifier,
        bool GenerateXmlDocument,
        bool SupportNullableReferenceTypes,
        bool SupportNullableStaticAnalysis)
    {
        internal static SourceGenerationOptions From(GeneratorExecutionContext context)
        {
            const string allowNullAttributeName = "System.Diagnostics.CodeAnalysis.AllowNullAttribute";
            var supportNullableStaticAnalysis = context.GetBuildGlobalOption(propertyName: nameof(SupportNullableStaticAnalysis), NullStaticAnalysisState.Default);
            var supportNullableReferenceTypes = context.Compilation.Options.NullableContextOptions is NullableContextOptions.Warnings or NullableContextOptions.Enable;

            return new(
                ConstructorAccessModifier: context.GetBuildGlobalOption(propertyName: nameof(ConstructorAccessModifier), AccessModifier.Public),
                GeneratedMethodsAccessModifier: context.GetBuildGlobalOption(propertyName: nameof(GeneratedMethodsAccessModifier), AccessModifier.Public),
                GenerateXmlDocument: context.GetBuildGlobalOption(propertyName: nameof(GenerateXmlDocument), true),
                SupportNullableReferenceTypes: supportNullableReferenceTypes,
                SupportNullableStaticAnalysis: supportNullableStaticAnalysis switch
                {
                    NullStaticAnalysisState.Enabled => true,
                    NullStaticAnalysisState.Disabled => false,
                    _ => context.Compilation is CSharpCompilation { LanguageVersion: >= LanguageVersion.CSharp8 } cs && cs.TypeByMetadataNameExists(allowNullAttributeName)
                }
            );
        }

        public string NullableReferenceSyntax => SupportNullableReferenceTypes ? "?" : string.Empty;
    }
}