﻿using System;
using MapTo.Extensions;
using static MapTo.Sources.Constants;

namespace MapTo.Sources
{
    internal static class MapRecordSource
    {
        internal static SourceCode Generate(MappingModel model)
        {
            using var builder = new SourceBuilder()
                .WriteLine(GeneratedFilesHeader)
                .WriteNullableContextOptionIf(model.Options.SupportNullableReferenceTypes)
                .WriteUsings(model.Usings)
                .WriteLine()

                // Namespace declaration
                .WriteLine($"namespace {model.Namespace}")
                .WriteOpeningBracket()

                // Class declaration
                .WriteLine($"partial record {model.TypeIdentifierName}")
                .WriteOpeningBracket();

            foreach (var targetSourceType in model.MappedSourceTypes)
            {
                if (targetSourceType.GenerateSecondaryConstructor)
                {
                    builder
                        .GenerateSecondaryConstructor(model)
                        .WriteLine();
                } 
            }

            // Class body
            

            builder
                .GeneratePrivateConstructor(model)
                
                .WriteLine()
                .GenerateFactoryMethod(model)

                // End class declaration
                .WriteClosingBracket()
                .WriteLine()

                // Extension class declaration
                .GenerateSourceTypeExtensionClass(model)
                

                // End namespace declaration
                .WriteClosingBracket();

            return new(builder.ToString(), $"{model.Namespace}.{model.TypeIdentifierName}.g.cs");
        }

        private static SourceBuilder GenerateSecondaryConstructor(this SourceBuilder builder, MappingModel model)
        {
            // grab first data from array

            foreach (var targetSourceType in model.MappedSourceTypes)
            {
                var sourceClassParameterName = targetSourceType.SourceTypeIdentifierName.ToCamelCase();
                if (model.Options.GenerateXmlDocument)
                {
                    builder
                        .WriteLine("/// <summary>")
                        .WriteLine($"/// Initializes a new instance of the <see cref=\"{model.TypeIdentifierName}\"/> class")
                        .WriteLine($"/// using the property values from the specified <paramref name=\"{sourceClassParameterName}\"/>.")
                        .WriteLine("/// </summary>")
                        .WriteLine($"/// <exception cref=\"ArgumentNullException\">{sourceClassParameterName} is null</exception>");
                }
                builder .WriteLine($"{model.Options.ConstructorAccessModifier.ToLowercaseString()} {model.TypeIdentifierName}({targetSourceType.SourceType} {sourceClassParameterName})")
                    .WriteLine($"    : this(new {MappingContextSource.ClassName}(), {sourceClassParameterName}) {{ }}");
            }



            return builder;

        }

        private static SourceBuilder GeneratePrivateConstructor(this SourceBuilder builder, MappingModel model)
        {
            const string mappingContextParameterName = "context";

            foreach (var targetSourceType in model.MappedSourceTypes)
            {
                var sourceClassParameterName = targetSourceType.SourceTypeIdentifierName.ToCamelCase();
                builder
                    .WriteLine(
                        $"private protected {model.TypeIdentifierName}({MappingContextSource.ClassName} {mappingContextParameterName}, {targetSourceType.SourceType} {sourceClassParameterName})")
                    .Indent()
                    .Write(": this(").WriteProperties(model, sourceClassParameterName, mappingContextParameterName)
                    .WriteLine(")")
                    .Unindent()
                    .WriteOpeningBracket()
                    .WriteLine($"if ({mappingContextParameterName} == null) throw new ArgumentNullException(nameof({mappingContextParameterName}));")
                    .WriteLine($"if ({sourceClassParameterName} == null) throw new ArgumentNullException(nameof({sourceClassParameterName}));")
                    .WriteLine()
                    .WriteLine($"{mappingContextParameterName}.{MappingContextSource.RegisterMethodName}({sourceClassParameterName}, this);")
                    .WriteClosingBracket();
            }
            // End constructor declaration
            return builder;
        }

        private static SourceBuilder WriteProperties(this SourceBuilder builder, MappingModel model, string sourceClassParameterName,
            string mappingContextParameterName)
        {

            foreach (var targetSourceType in model.MappedSourceTypes)
            {
                for (var i = 0; i < targetSourceType.SourceProperties.Length; i++)
                {
                    var property = targetSourceType.SourceProperties[i];
                    if (property.TypeConverter is null)
                    {
                        if (property.IsEnumerable)
                        {
                            builder.Write(
                                $"{property.Name}: {sourceClassParameterName}.{property.SourcePropertyName}.Select({mappingContextParameterName}.{MappingContextSource.MapMethodName}<{property.MappedSourcePropertyTypeName}, {property.EnumerableTypeArgument}>).ToList()");
                        }
                        else
                        {
                            builder.Write(property.MappedSourcePropertyTypeName is null
                                ? $"{property.Name}: {sourceClassParameterName}.{property.SourcePropertyName}"
                                : $"{property.Name}: {mappingContextParameterName}.{MappingContextSource.MapMethodName}<{property.MappedSourcePropertyTypeName}, {property.Type}>({sourceClassParameterName}.{property.SourcePropertyName})");
                        }
                    }
                    else
                    {
                        var parameters = property.TypeConverterParameters.IsEmpty
                            ? "null"
                            : $"new object[] {{ {string.Join(", ", property.TypeConverterParameters)} }}";

                        builder.Write(
                            $"{property.Name}: new {property.TypeConverter}().Convert({sourceClassParameterName}.{property.SourcePropertyName}, {parameters})");
                    }

                    if (i < targetSourceType.SourceProperties.Length - 1)
                    {
                        builder.Write(", ");
                    }
                }
            }

            

            return builder;
        }

        private static SourceBuilder GenerateFactoryMethod(this SourceBuilder builder, MappingModel model)
        {
            foreach (var targetSourceType in model.MappedSourceTypes)
            {
                var sourceClassParameterName = targetSourceType.SourceTypeIdentifierName.ToCamelCase();

                builder
                    .GenerateConvertorMethodsXmlDocs(model, sourceClassParameterName)
                    .WriteLineIf(model.Options.SupportNullableStaticAnalysis, $"[return: NotNullIfNotNull(\"{sourceClassParameterName}\")]")
                    .WriteLine(
                        $"{model.Options.GeneratedMethodsAccessModifier.ToLowercaseString()} static {model.TypeIdentifierName}{model.Options.NullableReferenceSyntax} From({targetSourceType.SourceType}{model.Options.NullableReferenceSyntax} {sourceClassParameterName})")
                    .WriteOpeningBracket()
                    .WriteLine(
                        $"return {sourceClassParameterName} == null ? null : {MappingContextSource.ClassName}.{MappingContextSource.FactoryMethodName}<{targetSourceType.SourceType}, {model.TypeIdentifierName}>({sourceClassParameterName});")
                    .WriteClosingBracket();
            }
            

            return builder;
        }

        private static SourceBuilder GenerateConvertorMethodsXmlDocs(this SourceBuilder builder, MappingModel model, string sourceClassParameterName)
        {
            if (!model.Options.GenerateXmlDocument)
            {
                return builder;
            }

            foreach (var targetSourceType in model.MappedSourceTypes)
            {
                builder
                    .WriteLine("/// <summary>")
                    .WriteLine($"/// Creates a new instance of <see cref=\"{model.TypeIdentifierName}\"/> and sets its participating properties")
                    .WriteLine($"/// using the property values from <paramref name=\"{sourceClassParameterName}\"/>.")
                    .WriteLine("/// </summary>")
                    .WriteLine($"/// <param name=\"{sourceClassParameterName}\">The instance of <see cref=\"{targetSourceType.SourceType}\"/> to use as source.</param>")
                    .WriteLine(
                        $"/// <returns>A new instance of <see cred=\"{model.TypeIdentifierName}\"/> -or- <c>null</c> if <paramref name=\"{sourceClassParameterName}\"/> is <c>null</c>.</returns>");
            }

            return builder;

        }

        private static SourceBuilder GenerateSourceTypeExtensionClass(this SourceBuilder builder, MappingModel model)
        {
            foreach (var targetSourceType in model.MappedSourceTypes)
            {
                builder
                    .WriteLine(
                        $"{model.Options.GeneratedMethodsAccessModifier.ToLowercaseString()} static partial class {targetSourceType.SourceTypeIdentifierName}To{model.TypeIdentifierName}Extensions")
                    .WriteOpeningBracket()
                    .GenerateSourceTypeExtensionMethod(model)
                    .WriteClosingBracket();
            }

            return builder;
        }

        private static SourceBuilder GenerateSourceTypeExtensionMethod(this SourceBuilder builder, MappingModel model)
        {

            foreach (var targetSourceType in model.MappedSourceTypes)
            {
                var sourceClassParameterName = targetSourceType.SourceTypeIdentifierName.ToCamelCase();

                builder
                    .GenerateConvertorMethodsXmlDocs(model, sourceClassParameterName)
                    .WriteLineIf(model.Options.SupportNullableStaticAnalysis, $"[return: NotNullIfNotNull(\"{sourceClassParameterName}\")]")
                    .WriteLine(
                        $"{model.Options.GeneratedMethodsAccessModifier.ToLowercaseString()} static {model.TypeIdentifierName}{model.Options.NullableReferenceSyntax} To{model.TypeIdentifierName}(this {targetSourceType.SourceType}{model.Options.NullableReferenceSyntax} {sourceClassParameterName})")
                    .WriteOpeningBracket()
                    .WriteLine($"return {sourceClassParameterName} == null ? null : new {model.TypeIdentifierName}({sourceClassParameterName});")
                    .WriteClosingBracket();

            }


            return builder;
        }
    }
}